import os  # for testing purposes
from dataclasses import dataclass
from typing import Optional

import risc_lasagna
import risc_lasagna.layer as layer


@dataclass
class DatabaseConfig:
    host: str
    port: int
    user: Optional[str] = None
    password: Optional[str] = None


@dataclass
class ApplicationConfig:
    app_host: str
    app_port: int
    database_config: DatabaseConfig


def load_application_config() -> ApplicationConfig:
    os.environ["DATABASE_CONFIG_HOST"] = "db-host"
    os.environ["DATABASE_CONFIG_PORT"] = "9876"
    return risc_lasagna.build(
        ApplicationConfig,
        [
            layer.EnvironmentLayer(),
            layer.DataClassDefaultLayer(
                ApplicationConfig("localhost", 12345, DatabaseConfig("localhost", 5432))
            ),
        ],
    )


if __name__ == "__main__":
    config = load_application_config()
    print(config)



import risc_lasagna.layer as layer
from risc_lasagna.tools import build, parse_value
from risc_lasagna.typedef import ConfigurationError

__all__ = [
    build,
    parse_value,
    ConfigurationError,
    layer,
]



import os  # for testing purposes
from dataclasses import dataclass

import risc_lasagna
import risc_lasagna.layer as layer


@dataclass
class ApplicationConfig:
    app_host: str
    app_port: int


def load_application_config() -> ApplicationConfig:
    os.environ["APP_HOST"] = "remote_host"
    return risc_lasagna.build(
        ApplicationConfig,
        [
            layer.EnvironmentLayer(),
            layer.DataClassDefaultLayer(ApplicationConfig("localhost", 12345)),
        ],
    )


if __name__ == "__main__":
    config = load_application_config()
    print(config)



from pathlib import Path

from risc_lasagna.layer.dict_layer import DictLayer
from risc_lasagna.typedef import ConfigurationError


class _DotEnvLayer(DictLayer):
    pass


def DotEnvLayer(
    file: Path,
    raise_exception_if_file_not_readable: bool = False,
    comment_char: str = "#",
    capitalize_keys: bool = True,
    variable_prefix: str = "",
) -> _DotEnvLayer:
    """
    Reads a .env files and parses it into an environment layer. Example .env
    EFRE_RESINET_A=123456789
    EFRE_RESINET_B=123456789

    Creates the DotEnv Layer from a given file
    :param file: Path to file
    :param raise_exception_if_file_not_readable: If the layer should raise an exception if file can not be read
    or act as an 'empty' layer
    :param comment_char: First non whitespace char which marks the line as a comment (will be ignored)
    :param capitalize_keys: If keys should automatically be capitalized -> makes everything case insensitive. (makes
    sense in some environments)
    :param variable_prefix: Prefix which is prepended before every key
    """
    map_ = {}
    try:
        with open(file, "r") as f:
            # open and parse the file, ignore comment lines (starting with comment char)
            for line in f.readlines():
                if not line.strip().startswith(comment_char):
                    line = tuple(line.split("="))
                    # skip empty lines or comments
                    if len(line) != 2:
                        continue
                    # add new keys to the map
                    map_[line[0].strip()] = line[1].strip()
    except (IOError, OSError) as e:
        if raise_exception_if_file_not_readable:
            raise ConfigurationError from e

    return _DotEnvLayer(
        map_, variable_prefix=variable_prefix, capitalize_keys=capitalize_keys
    )



import tempfile  # for testing purposes
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional

import risc_lasagna
import risc_lasagna.layer as layer


@dataclass
class DatabaseConfig:
    host: str
    port: int
    user: Optional[str] = None
    password: Optional[str] = None


@dataclass
class HealthCheck:
    rate_in_ms: int
    url: str


@dataclass
class ApplicationConfig:
    host: str
    port: int
    elements: List[str]
    health_check: Optional[HealthCheck] = field(default=None)
    protocol: str = field(default="https")
    database_configs: Optional[List[DatabaseConfig]] = field(default=None)


__example_yaml = """
host: remote_host
health_check:
  rate_in_ms: 30
  url: localhost:6543
elements:
  - abc
  - def
  - ghi
database_configs:
  - host: remote_db
    port: 1234
  - host: remote_db
    port: 1234
    user: postgres
    password: super_safe_password
"""


def load_application_config() -> ApplicationConfig:
    # set the environment variable with a prefix
    with tempfile.NamedTemporaryFile() as temp:
        temp.write(bytes(__example_yaml, encoding="utf-8"))
        temp.flush()
        return risc_lasagna.build(
            ApplicationConfig,
            [
                layer.YamlLayer(Path(temp.name)),
                layer.DataClassDefaultLayer(
                    ApplicationConfig(
                        "localhost",
                        12345,
                        ["a", "b", "c"],
                        database_configs=[DatabaseConfig("localhost", 5432)],
                    )
                ),
            ],
        )


if __name__ == "__main__":
    config = load_application_config()
    print(config)



import os
from typing import Callable, Optional

from risc_lasagna.helper import normalize_variable_prefix
from risc_lasagna.layer.dict_layer import DictLayer


class _EnvironmentLayer(DictLayer):
    pass


def EnvironmentLayer(
    variable_prefix: Optional[str] = None,
    capitalize_keys: bool = True,
    normalize_prefix: Callable[[str], str] = normalize_variable_prefix,
):
    """
    Provides access to os.environ keys. Filters the environment by a given variable_prefix if given.
    """
    """
    Creates an environment layer which can be scorped by a variable prefix.
    :param variable_prefix: The variable prefix which is added to every key
    :param capitalize_keys: If keys should be capitalized for access
    """
    # filter for the prefix (avoids importing whole environment)
    if variable_prefix is not None:
        map_ = {
            key: os.environ[key]
            for key in os.environ.keys()
            if key.startswith(variable_prefix)
        }
    else:
        map_ = dict(os.environ)

    if capitalize_keys:
        map_ = {key.upper(): value for key, value in map_.items()}

    return _EnvironmentLayer(
        map_,
        variable_prefix=variable_prefix,
        capitalize_keys=capitalize_keys,
        normalize_prefix=normalize_prefix,
    )



import logging
import os
from collections.abc import MutableMapping
from pathlib import Path

import yaml

from risc_lasagna.layer.dict_layer import DictLayer
from risc_lasagna.typedef import ConfigurationError

logger = logging.getLogger(__name__)


class _YamlLayer(DictLayer):
    pass


def YamlLayer(
    file: Path,
    raise_exception_if_file_not_readable: bool = False,
    variable_prefix: str = "",
) -> _YamlLayer:
    """
    Creates the YAML Layer from a given file
    :param file: Path to file
    :param raise_exception_if_file_not_readable: If the layer should raise an exception if file can not be read
    or act as an 'empty' layer
    :param variable_prefix: Prefix which is prepended before every key
    """
    if raise_exception_if_file_not_readable and not os.access(file, os.R_OK):
        raise ConfigurationError from ValueError(
            f"File {file} is not readable or does not exist!"
        )

    map_ = {}
    try:
        with open(file, "r") as f:
            # open and parse the file, ignore comment lines (starting with comment char)
            parsed = yaml.safe_load(f)
            map_ = _flatten(parsed)
    except (IOError, yaml.YAMLError) as e:
        logger.error(f"Error reading yml file {file}", exc_info=e)
        if raise_exception_if_file_not_readable:
            raise ConfigurationError from e

    # yaml layer are never capitalized because yaml is case sensitive by default
    return _YamlLayer(map_, capitalize_keys=False, variable_prefix=variable_prefix)


def _flatten(d, parent_key="", sep="_"):
    items = []
    for key, value in d.items():
        # define new key
        new_key = f"{parent_key}{sep}{key}" if parent_key else key

        if isinstance(value, MutableMapping):
            # recursively parse sub dict
            items.extend(_flatten(value, new_key, sep=sep).items())
        elif isinstance(value, list):
            # parse lists and their sub elements
            values = []
            for element in value:
                # parse sub dicts and flatten them
                if isinstance(element, MutableMapping):
                    element = _flatten(element, "", sep=sep)

                # add normal values back to list
                values.append(element)

            # add simple lists (e.g. no objects) to key
            if len(values) > 0:
                items.append((new_key, values))
        else:
            # case for simple key
            items.append((new_key, value))
    return dict(items)



import dataclasses
import typing
from dataclasses import MISSING, Field, fields, is_dataclass
from typing import Any, List, Union

from .helper import is_list, is_optional, normalize_variable_prefix
from .layer import AggregationLayer, DictLayer
from .layer.typedef import Layer


def parse_value(type_: type, value: Any):
    """
    Parse values which should be simple types. Handles some edge caes (actually parses bool 'f' as false, ...)
    :param type_: The type to which it is parsed to
    :param value: The value (likely string)
    :return: The parsed value
    """
    # edge cases for boolean parsing
    if type_ == bool and value in ["f", "F", "False", "false", "", 0, "0", "None"]:
        return False

    # trust that complex types are initiated correctly
    if (
        typing.get_origin(type_) is typing.Union
        or typing.get_origin(type_) is typing.Optional
    ):
        return value

    # in case it is not an object but a class (for class methods, grouping) we do not
    # need to instantiate it
    if typing.get_origin(type_) is typing.get_origin(typing.Type):
        return value

    # default
    return type_(value)


def build(data_class, layers: Union[List[Layer], Layer], variable_prefix: str = ""):
    """
    Parse config from a given layer into a data class
    :param data_class: The data class that should be created
    :param layers: A layer (or list of layer) which contains the necessary values
    :param variable_prefix: A prefix which ALL variables have
    :raise ValueError: If parameter data_class is not a data class according to the dataclasses.is_dataclass function
    :raise ValueError: If required field in the data_class is not set in the layer
    :raise ValueError: If the annotated type in the data_class can not be used to parse the provided data from the layer
    :return: The parsed data class from the layer
    """
    if not is_dataclass(data_class):
        raise ValueError("Please provide a dataclass that represents the config")

    # convert list of layer to aggregation layer
    if isinstance(layers, list):
        layers = AggregationLayer(layers)

    # build dict for constructor
    dict_for_constructor = {}
    data_field: Field
    # read data from layer and construct dict for data class constructor
    for data_field in fields(data_class):
        # build the name in the environment (with prefix, ...)
        name_in_environment = (
            f"{normalize_variable_prefix(variable_prefix)}{data_field.name}"
        )

        # extract type from optional wrapper
        suggested_type = data_field.type
        if is_optional(data_field.type):
            suggested_type = typing.get_args(data_field.type)[0]

        # parse nested dataclasses
        if is_dataclass(suggested_type):
            parsed_nested = build(
                suggested_type, layers, variable_prefix=name_in_environment
            )
            dict_for_constructor = {
                **dict_for_constructor,
                **{data_field.name: parsed_nested},
            }
        else:
            # value is missing, which is fine. When the value is required (no default set) the constructor
            # throws an error
            if not layers.contains(name_in_environment):
                if data_field.default is MISSING:
                    raise ValueError(
                        f"The required attribute '{data_field.name}' is not set, "
                        f"expected the variable '{name_in_environment}'"
                    )

            # parse the value using the type information in the data field
            value = layers.get(name_in_environment)
            if value is None:
                value = data_field.default

            if value is not None and value is not MISSING:
                try:
                    if is_list(suggested_type):
                        # get type args of list, can be a primitive, a dataclass or optional
                        args = typing.get_args(suggested_type)

                        # no type specified -> we can only try to guess it later
                        suggested_type = None
                        if len(args) > 0:
                            suggested_type = args[0]

                        values = []
                        for element in layers.get(name_in_environment):
                            # recursively parse nested class
                            if is_dataclass(suggested_type or type(element)):
                                # parse default data classes (if nested into config in some way)
                                if is_dataclass(element):
                                    element = dataclasses.asdict(element)

                                parsed_nested = build(
                                    suggested_type,
                                    DictLayer(element, capitalize_keys=False),
                                )
                                values.append(parsed_nested)
                            else:
                                # try to parse value from inferred type
                                values.append(parse_value(type(element), element))

                        value = values
                    else:
                        value = parse_value(suggested_type, value)
                except ValueError as error:
                    raise ValueError(
                        f"Could not parse data field '{data_field.name}' of data class '{data_class.__name__}' "
                        f"(full name in layer: '{name_in_environment}'). Please check if the value or "
                        f"type of the data field is set correctly."
                    ) from error

            # in case the given variable is not in the environment the value is of type MISSING
            # if the value is not set in other layer (in case of an aggregation layer) the constructor should
            # fail!
            if value is not MISSING:
                dict_for_constructor = {
                    **dict_for_constructor,
                    **{data_field.name: value},
                }

    return data_class(**dict_for_constructor)



import os  # for testing purposes
from dataclasses import dataclass

import risc_lasagna
import risc_lasagna.layer as layer


@dataclass
class ApplicationConfig:
    app_host: str
    app_port: int


def load_application_config() -> ApplicationConfig:
    # add environment for testing, this should be set by your deployment tool
    os.environ["APP_HOST"] = "localhost"
    os.environ["APP_PORT"] = "12345"
    return risc_lasagna.build(ApplicationConfig, layer.EnvironmentLayer())


if __name__ == "__main__":
    config = load_application_config()
    print(config)



from risc_lasagna.layer.aggregation_layer import AggregationLayer
from risc_lasagna.layer.dataclass_default_layer import DataClassDefaultLayer
from risc_lasagna.layer.dict_layer import DictLayer
from risc_lasagna.layer.dotenv_layer import DotEnvLayer
from risc_lasagna.layer.environment_layer import EnvironmentLayer
from risc_lasagna.layer.yaml_layer import YamlLayer

__all__ = [
    DictLayer,
    AggregationLayer,
    DataClassDefaultLayer,
    DotEnvLayer,
    EnvironmentLayer,
    YamlLayer,
]



import logging
import os
from typing import List

from risc_lasagna.helper import print_dict_like_list
from risc_lasagna.layer.dotenv_layer import _DotEnvLayer
from risc_lasagna.layer.environment_layer import _EnvironmentLayer
from risc_lasagna.layer.typedef import Layer
from risc_lasagna.layer.yaml_layer import _YamlLayer
from risc_lasagna.typedef import ConfigurationError

logger = logging.getLogger(__name__)


class AggregationLayer(Layer):
    """
    This layer accept a list of layer. These then act as a single 'layer'. The ordering decides which variable
    from which layer takes priority. (first in the list overwrites all the ones after that).
    """

    def __init__(self, layers: List[Layer]):
        """
        :param layers: A list of layer which are used to lookup values. Layer are consulted in the given order.
        Meaning the first layer overwrites any layer after it.
        """
        layer_types = [type(layer) for layer in layers]
        if _DotEnvLayer in layer_types and _YamlLayer in layer_types:
            raise ConfigurationError(
                "DotEnv and Yaml layer are not compatible! Please use one or the other."
            )
        if _YamlLayer in layer_types and _EnvironmentLayer in layer_types:
            logger.info(
                "Yaml layer and environment layer are only partially compatible. "
                "Lists can not be defined via environment variables and yaml is case sensitive while the environment "
                "is not by default!"
            )

        self._layers = layers

    def get(self, key):
        for layer in self._layers:
            value = layer.get(key)
            if value is not None:
                return value
        return None

    def keys(self) -> List[str]:
        keys = set()
        for layer in self._layers:
            keys.update(layer.keys())
        return list(keys)

    def contains(self, key: str) -> bool:
        for layer in self._layers:
            if layer.contains(key):
                return True
        return False

    def __repr__(self):
        """
        Output the aggregated layer and a list of internal layer.
        """
        # show computed what the layer contains
        map_ = {key: self.get(key) for key in self.keys()}
        repr_ = print_dict_like_list(map_, title=self.__class__.__name__)

        repr_ = f"{repr_}{os.linesep}... based on ...{os.linesep * 2}"

        # show each layer
        for index, layer in enumerate(self._layers, start=-len(self._layers)):
            repr_ = f"{repr_}Layer {abs(index)} {repr(layer)}"
        return repr_



import os
import typing
from dataclasses import MISSING, Field, fields, is_dataclass
from enum import Enum
from typing import Any, Dict, Optional, Union

from risc_lasagna.typedef import ConfigurationError


def print_dict_like_list(
    map_: Dict[str, Any], indent: int = 4, title: Optional[str] = None
):
    """
    Prints a dict in a list like form (recursively).
    :param map_: The dict
    :param indent: How much intent to add before keys
    :param title: A title which is added at the beginning
    :return: A string representing the dict
    # >>> print_dict_like_list({'A': 'a', 'B': 123, 'C': [1,2,3]})
    # '    A=\n    B=123\n    C=[1, 2, 3]\n'
    # >>> print_dict_like_list({'A': 'a'}, title='test')
    # 'test\n    A=a\n'
    # >>> print_dict_like_list({'a': 'a'})
    # 'a=a'
    """
    if title is not None:
        repr_ = f"{title}:{os.linesep}"
    else:
        repr_ = ""

    for key in map_:
        # recursively print the inner dicts
        if isinstance(map_[key], dict):
            repr_ = f'{repr_}{" " * indent}{key}:{os.linesep}{print_dict_like_list(map_[key], indent=indent * 2)}'
            continue
        # print the the key
        repr_ = f'{repr_}{" " * indent}{key}={map_[key]}{os.linesep}'
    return repr_


def normalize_variable_prefix(prefix: str) -> str:
    """
    Normalizes the prefix to contain an '_' at the end
    :param prefix: The prefix which should be normalized
    :return: The prefix which ends in '_' or an empty prefix
    >>> normalize_variable_prefix('')
    ''
    >>> normalize_variable_prefix('PREFIX')
    'PREFIX_'
    >>> normalize_variable_prefix('PREFIX_')
    'PREFIX_'
    """
    if prefix == "":
        return prefix
    return prefix if prefix.endswith("_") else f"{prefix}_"


def prepend_variable_prefix(prefix: str, variable: str) -> str:
    """
    Appends the prefix to a string variable if it does not already start with it
    :param prefix:
    :param variable:
    :return:
    >>> prepend_variable_prefix('TEST_', 'VAR')
    'TEST_VAR'
    >>> prepend_variable_prefix('TEST_', 'TEST_VAR')
    'TEST_VAR'
    >>> prepend_variable_prefix('', 'VAR')
    'VAR'
    >>> prepend_variable_prefix('', '')
    ''
    """
    if variable.startswith(prefix):
        return variable
    return f"{prefix}{variable}"


class ConfigStyle(Enum):
    DOTENV = "DOTENV"
    YAML = "YAML"


def print_expected_variables_names(
    prefix: str, data_class, style: ConfigStyle = ConfigStyle.DOTENV
):
    if not is_dataclass(data_class):
        raise ConfigurationError(
            f"data_class needs to be a data class! It is a {type(data_class)}"
        )

    if style == ConfigStyle.DOTENV:
        field: Field
        prefix = normalize_variable_prefix(prefix)
        for field in fields(data_class):
            if is_dataclass(field.type):
                # recursively parse the inner dataclass, flattens it out
                print_expected_variables_names(
                    f"{prefix}{field.name.upper()}", field.type
                )
                continue
            if is_list(field.type):
                raise ConfigurationError("Lists are not supported in dot env style")

            # add defaults to output
            if field.default is MISSING:
                default = f"<{field.type.__name__}>"
            else:
                default = field.default

            print(f"{prefix}{field.name.upper()}={default}")
    elif style == ConfigStyle.YAML:
        raise ConfigurationError("Printing yaml style is currently not supported!")


def is_optional(type_hint):
    """
    Checks if a given type is typing.Optional
    >>> from typing import List
    >>> is_optional(Optional[str])
    True
    >>> is_optional(Optional[List[str]])
    True
    >>> is_optional(List[str])
    False
    >>> is_optional(str)
    False
    """
    return typing.get_origin(type_hint) is Union and type(None) in typing.get_args(
        type_hint
    )


def is_optional_list(type_hint) -> bool:
    """
    Checks if a given type hint is an optional list.
    >>> from typing import List
    >>> is_optional_list(List[str])
    False
    >>> is_optional_list(Optional[List[int]])
    True
    >>> is_optional_list(Optional[List])
    True
    """
    return is_optional(type_hint) and list in [
        typing.get_origin(arg) for arg in typing.get_args(type_hint)
    ]


def is_list(type_hint) -> bool:
    """
    Checks if a given type hint is a list. Does not matter if it is optional or the builtin list type.
    >>> from typing import List
    >>> is_list(list)
    True
    >>> is_list(List)
    True
    >>> is_list(List[str])
    True
    >>> is_list(Optional[List])
    True
    >>> is_list(Optional[List[str]])
    True
    >>> is_list(Optional[str])
    False
    """
    return (
        type_hint == list
        or is_optional_list(type_hint)
        or typing.get_origin(type_hint) == list
    )



import typing
from dataclasses import Field, fields, is_dataclass
from inspect import isclass
from typing import Any, Callable, Dict, List, Optional

from risc_lasagna.helper import is_list, is_optional, print_dict_like_list
from risc_lasagna.layer.dict_layer import DictLayer
from risc_lasagna.layer.typedef import Layer
from risc_lasagna.typedef import ConfigurationError


class DataClassDefaultLayer(Layer):
    """
    This class maps a dataclass instance to an environment. Variables are expanded in a dict and can
    be accessed by a key.
    Examples:
    """

    def __init__(
        self,
        dataclass_instance,
        variable_prefix: str = None,
        capitalize_keys: bool = True,
        dict_layer_generator: Callable[
            [Dict[str, Any]], DictLayer
        ] = lambda map_: DictLayer(map_),
    ):
        if not is_dataclass(dataclass_instance) or isclass(dataclass_instance):
            raise ConfigurationError from ValueError(
                "Variable dataclass must be a valid instance of a dataclass!"
            )

        # translate db result to dataclass names. The resulting dict is then used in the constructor
        _map = DataClassDefaultLayer.__build_map_from(
            dataclass_instance, variable_prefix, capitalize_keys
        )
        self._dict_layer = dict_layer_generator(_map)

    def get(self, key):
        return self._dict_layer.get(key)

    def keys(self) -> List[str]:
        return self._dict_layer.keys()

    def contains(self, key: str) -> bool:
        return self._dict_layer.contains(key)

    def __repr__(self):
        return print_dict_like_list(self._dict_layer.map, title=self.__class__.__name__)

    @staticmethod
    def __build_with_prefix(
        key: str, variable_prefix: Optional[str], capitalize_keys: bool
    ):
        if variable_prefix is not None:
            key = f"{DataClassDefaultLayer.__normalize_prefix(variable_prefix)}{key}"

        return key.upper() if capitalize_keys else key

    @staticmethod
    def __normalize_prefix(variable_prefix: Optional[str]) -> str:
        if variable_prefix is None:
            return ""
        return (
            variable_prefix if variable_prefix.endswith("_") else f"{variable_prefix}_"
        )

    @staticmethod
    def __build_map_from(
        dataclass_instance, variable_prefix: Optional[str], capitalize_keys: bool
    ) -> Dict[str, Any]:
        field: Field
        mapped_and_flattened_dict = {}
        for field in fields(dataclass_instance):
            type_ = field.type
            if is_optional(type_) or is_list(type_):
                type_ = typing.get_args(type_)[0]

            if is_dataclass(type_):
                # recursively parse the inner dataclass, flattens it out
                mapped_and_flattened_dict = {
                    **mapped_and_flattened_dict,
                    **DataClassDefaultLayer.__build_map_from(
                        getattr(dataclass_instance, field.name),
                        f"{DataClassDefaultLayer.__normalize_prefix(variable_prefix)}{field.name}",
                        capitalize_keys,
                    ),
                }
                continue
            mapped_and_flattened_dict[
                DataClassDefaultLayer.__build_with_prefix(
                    field.name, variable_prefix, capitalize_keys
                )
            ] = getattr(dataclass_instance, field.name)
        return mapped_and_flattened_dict



from abc import abstractmethod
from typing import List, Protocol


class Layer(Protocol):
    @abstractmethod
    def get(self, key):
        pass

    @abstractmethod
    def keys(self) -> List[str]:
        pass

    @abstractmethod
    def contains(self, key: str) -> bool:
        pass

    @abstractmethod
    def __repr__(self):
        pass



from dataclasses import dataclass

import pytest

from risc_lasagna.layer.dataclass_default_layer import DataClassDefaultLayer
from risc_lasagna.typedef import ConfigurationError


@dataclass
class NestedExample:
    nested_a: str


@dataclass
class Example:
    a: str
    b: str
    c: NestedExample


class NotADataClass:
    a: str


@pytest.mark.parametrize(
    "key,value",
    [("PROJECT_A", "a"), ("PROJECT_B", "b"), ("PROJECT_C_NESTED_A", "nested_a")],
)
def test_default_nested(key: str, value: str):
    layer = DataClassDefaultLayer(
        Example("a", "b", NestedExample("nested_a")), variable_prefix="PROJECT"
    )
    assert key in layer.keys()
    assert layer.get(key) == value


def test_not_a_dataclass():
    with pytest.raises(ConfigurationError):
        DataClassDefaultLayer(NotADataClass())


def test_keys():
    layer = DataClassDefaultLayer(
        Example("a", "b", NestedExample("nested_a")), variable_prefix="PROJECT"
    )
    assert len(layer.keys()) == 3



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['risc_lasagna', 'risc_lasagna.examples', 'risc_lasagna.layer']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0']

setup_kwargs = {
    'name': 'risc-lasagna',
    'version': '2.0.0',
    'description': 'A package that makes it easy to load configuration variables from multiple sources and build a single config which depends on them.',
    'long_description': '[![coverage report](https://gitdma.risc-software.at/risc_ds/lasagna/badges/master/coverage.svg)](https://gitdma.risc-software.at/risc_ds/lasagna/-/commits/master)\n[![pipeline status](https://gitdma.risc-software.at/risc_ds/lasagna/badges/master/pipeline.svg)](https://gitdma.risc-software.at/risc_ds/lasagna/-/commits/master) \n\n# Lasagna\n\nA package that makes it easy to load configuration variables from \nmultiple sources and build a single config data class.\n\n## The Idea\nThe idea is to \'stack\' the sources like lasagna and read them from the \nbottom up. This means the layer at the bottom can set defaults and any \nlayers on top of it overwrite these defaults. Which allows to build \nconfigurations from multiple sources with one clear end result.\n\nThe configuration must be wrapped up in a single data class. \nThis enables the use of type hints as well as taking advantage of any other standard tooling\nwhich operates on data classes. The library also supports nested data classes. This results in \nmultiple *advantages*:\n\n- When refactoring, for example adding a required field to the dataclass, \nwithout adding it to the default config, results in a error on startup.\n- Type annotations can support the developer. No more guessing what is read.\n- Automatically parsing simple types from configuration files and the environment\n- No boilerplate for accessing dicts / ...\n- Easily combine multiple sources for the environment **(data class instances, .env files, environment)**\n\n## Setup\n\nNo additional packages are currently required to use `lasagna`. Simply install it.\n```shell script\npip install risc-lasagna\n```\n\n## Development\n\nFor development simply clone the repository. To release a version change the version specifier\nin the `setup.cfg` and commit the changes with a tag. The tag name should be `<VERSION_IN_SETUP_CFG>` (e.g. 2.0.2).\n\n## Tests\n\nSee task in `pyproject.toml`.\n\n## Examples\n\nThe following section contains a few examples whiche illustrate how to use the package.\n\n### Simple Example\nTo build your config you need to first specify a dataclass which \nencompasses the necessary values. The next step is to specify where the\nvalues come from. A simple example is to use the `EnvironmentLayer`. Which \nreads all values from the environment and loads them in the dataclass.\n\n```python\nfrom dataclasses import dataclass\nimport os  # for testing purposes\n\nimport risc_lasagna as lasagna\nimport risc_lasagna.layer as layer\n\n\n@dataclass\nclass ApplicationConfig:\n    app_host: str\n    app_port: int\n\n\ndef load_application_config() -> ApplicationConfig:\n    # add environment for testing, this should be set by your deployment tool\n    os.environ[\'APP_HOST\'] = \'localhost\'\n    os.environ[\'APP_PORT\'] = \'12345\'\n    return lasagna.build(ApplicationConfig, layer.EnvironmentLayer())\n```\nExecuting this example results in a simple parsed application config\n```python\n>>> load_application_config()\nApplicationConfig(app_host=\'localhost\', app_port=12345)\n```\n\n### Simple Example with default layer\nSometimes it is easier to provide \'defaults\' for every dataclass. This can be accomplished by providing\na default layer using your dataclass. The same example as before can be rewritten to use a default \nlayers using the `DataClassDefaultLayer` and the `AggregationLayer`. The `AggregationLayer` \nSimply combines the layers from top to bottom. The bottom most layer is overwritten\nby every the top layers. \n\n```python\nfrom dataclasses import dataclass\nimport os  # for testing purposes\n\nimport risc_lasagna as lasagna\nimport risc_lasagna.layer as layer\n\n\n@dataclass\nclass ApplicationConfig:\n    app_host: str\n    app_port: int\n\n\ndef load_application_config() -> ApplicationConfig:\n    os.environ[\'APP_HOST\'] = \'remote_host\'\n    return lasagna.build(ApplicationConfig, \n        [\n            layer.EnvironmentLayer(),\n            layer.DataClassDefaultLayer(\n                ApplicationConfig(\'localhost\', 12345)\n            ),\n        ],\n    )\n```\nRunning the example results in the `EnvironmentLayer` overwriting the `APP_HOST` variable of the\n`DataClassDefaultLayer`.\n```python\n>>> load_application_config()\nApplicationConfig(app_host=\'remote_host\', app_port=12345)\n```\n\n### Example with nested dataclass\nFor larger configurations it might be easier or convenient \nto assemble it from multiple dataclasses. This is possible by \nprefixing the existing variables in the environment with the name of \nthe property. Here is an example:\n```python\nfrom dataclasses import dataclass\nimport os  # for testing purposes\nfrom typing import Optional\n\nimport risc_lasagna as lasagna\nimport risc_lasagna.layer as layer\n\n\n@dataclass\nclass DatabaseConfig:\n    host: str\n    port: int\n    user: Optional[str] = None\n    password: Optional[str] = None\n\n\n@dataclass\nclass ApplicationConfig:\n    app_host: str\n    app_port: int\n    database_config: DatabaseConfig  # the name of this variable is used as a prefix\n\n\ndef load_application_config() -> ApplicationConfig:\n    # set the environment variable with a prefix\n    os.environ[\'DATABASE_CONFIG_HOST\'] = \'db-host\'\n    return lasagna.build(\n        ApplicationConfig,\n        [\n            layer.EnvironmentLayer(),\n            layer.DataClassDefaultLayer(\n                ApplicationConfig(\'localhost\', 12345, DatabaseConfig(\'localhost\', 5432))\n            ),\n        ],\n    )\n```\n\nThe example results in the `EnvironmentLayer` overwriting the defaults:\n\n```python\n>>> load_application_config()\nApplicationConfig(app_host=\'localhost\', app_port=12345, database_config=DatabaseConfig(host=\'db-host\', port=5432, user=None, password=None))\n```\n\n### Example with YAML and lists\nFor more complex configuration YAML is a better option. Yaml also supports lists and objects inside lists.\n>HINT: If you use lists with nested dataclasses be aware that only a single type per list is supported.\n\n```python\nimport tempfile  # for testing purposes\nfrom dataclasses import dataclass, field\nfrom pathlib import Path\nfrom typing import Optional, List\n\nimport risc_lasagna as lasagna\nimport risc_lasagna.layer as layer\n\n\n@dataclass\nclass DatabaseConfig:\n    host: str\n    port: int\n    user: Optional[str] = None\n    password: Optional[str] = None\n\n\n@dataclass\nclass HealthCheck:\n    rate_in_ms: int\n    url: str\n\n\n@dataclass\nclass ApplicationConfig:\n    host: str\n    port: int\n    elements: List[str]\n    health_check: Optional[HealthCheck] = field(default=None)\n    protocol: str = field(default=\'https\')\n    database_configs: Optional[List[DatabaseConfig]] = field(default=None)\n\n\n__example_yaml = """\nhost: remote_host\nhealth_check:\n  rate_in_ms: 30\n  url: localhost:6543\nelements:\n  - abc\n  - def\n  - ghi\ndatabase_configs:\n  - host: remote_db\n    port: 1234\n  - host: remote_db\n    port: 1234\n    user: postgres\n    password: super_safe_password\n"""\n\n\ndef load_application_config() -> ApplicationConfig:\n    # set the environment variable with a prefix\n    with tempfile.NamedTemporaryFile() as temp:\n        temp.write(bytes(__example_yaml, encoding=\'utf-8\'))\n        temp.flush()\n        return lasagna.build(\n            ApplicationConfig,\n            [\n                layer.YamlLayer(Path(temp.name)),\n                layer.DataClassDefaultLayer(\n                    ApplicationConfig(\n                        \'localhost\', 12345, [\'a\', \'b\', \'c\'],\n                        database_configs=[DatabaseConfig(\'localhost\', 5432)]\n                    )\n                ),\n            ],\n        )\n```\n\nExecuting the example results in this dataclass:\n```\n>>> load_application_config() \nApplicationConfig(host=\'remote_host\', port=12345, elements=[\'abc\', \'def\', \'ghi\'], health_check=HealthCheck(rate_in_ms=30, url=\'localhost:6543\'), protocol=\'https\', database_configs=[DatabaseConfig(host=\'remote_db\', port=1234, user=None, password=None), DatabaseConfig(host=\'remote_db\', port=1234, user=\'postgres\', password=\'super_safe_password\')])\n```\n\n### FAQ\n\n#### What happens if not enough variables are present to instantiate the data class?\nIn this case python raises an Error when the constructor is called (\'... missing required attribute ...\'). There are\ntwo options to avoid the error:\n1. Add the necessary variable to the environment\n2. Make the attribute optional\n\n#### I have a global prefix (\'RISCSW\') which is prepended before any variable, do i need to add this to my variable names?\nNo. The `build` methods defines a `variable_prefix` parameter which can be used to deal with such cases.\nOther layers may also include a `variable_prefix` which is then applied to all of the variables in the layer.\n\n#### I want to read .env files. Is this possible?\nYes. You can use the `DotEnvLayer` for this. This layer also supports comment in the .env file.\nThe file path has to be provided - Either make it fixed or write it to another environment variable, to set it\ndynamically.\n\n#### I want to read .yml files. Is this possible?\nYes. Use the YamlLayer. This has a few limitations (directly nested lists are not supported). \nBut works nicely for most cases. \n\n#### I want to read complex types (such as np.array, Unions)\nCurrently only simple types are supported. If you want to contribute contact the maintainer. Optional is the only \nsupported type.\n',
    'author': 'Dominik Falkner',
    'author_email': 'dominik.falkner@risc-software.at',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Callable, Dict, List, Optional

from risc_lasagna.helper import (
    normalize_variable_prefix,
    prepend_variable_prefix,
    print_dict_like_list,
)
from risc_lasagna.layer.typedef import Layer


class DictLayer(Layer):
    """
    A very simple layer for loading dict configurations. Uses a dict as for it's basis.
    """

    def __init__(
        self,
        map_: Dict[str, any],
        variable_prefix: Optional[str] = None,
        capitalize_keys: bool = True,
        normalize_prefix: Callable[[str], str] = normalize_variable_prefix,
    ):
        self.map = map_
        self._variable_prefix = variable_prefix
        self._capitalize_keys = capitalize_keys
        self._normalize_prefix = normalize_prefix

    def get(self, key: str):
        """
        Searches for the key in the dict. Capitalizes the key if capitalize_keys is set to true in the constructor.
        Adds the prefix if variable_prefix is set in the constructor, the prefix is normalized with the
        normalize_prefix function.
        :param key: The string key
        :return: Returns the value of key if set otherwise None
        """
        return self.map.get(self._normalize_key(key), None)

    def keys(self) -> List[str]:
        """
        :return: Returns list of available keys
        """
        return list(self.map.keys())

    def _normalize_key(self, key: str) -> str:
        if self._capitalize_keys:
            key = key.upper()
        if self._variable_prefix is not None and not key.startswith(
            self._normalize_prefix(self._variable_prefix)
        ):
            # add prefix if not already added to key
            key = prepend_variable_prefix(
                self._normalize_prefix(self._variable_prefix), key
            )
        return key

    def contains(self, key: str) -> bool:
        return self._normalize_key(key) in self.map.keys()

    def __repr__(self):
        """
        :return: the representation of the layer
        """
        return print_dict_like_list(self.map, title=self.__class__.__name__.lstrip("_"))
